import os
import git
from distutils.dir_util import copy_tree
import subprocess
import shutil

from jinja2 import Environment, FileSystemLoader

from ..file_io import dump, load, dump_to_string, get_file_basename, make_dir, get_extension
from .process_template import process_template

class DocumentFile:
  def __init__(self, path):
    self.path = os.path.abspath(path)
    self.file_base_name = get_file_basename(path, depth=2)
    self.markdown_output_path = None
    self.corresponding_schema_path = None

  def set_markdown_output_path(self, path):
    self.markdown_output_path = os.path.abspath(path)

  def set_corresponding_schema_path(self, path):
    self.corresponding_schema_path = os.path.relpath(path)


class HugoWeb:
  def __init__(self, docs_source_directory, build_directory, schema_directory=None):
    self.build_directory = os.path.abspath(build_directory)
    self.docs_source_directory = os.path.abspath(docs_source_directory)
    self.docs_config_directory = os.path.join(self.docs_source_directory,"web")
    if schema_directory is None:
      self.source_schema_directory_path = os.path.abspath(os.path.join(self.docs_source_directory,os.pardir,"schema"))
    self.title = os.path.relpath(self.docs_source_directory)
    self.description = ""
    self.author = ""
    self.has_logo = False
    self.colors = {
      "primary": "#30638E",
      "secondary": "#FFA630",
      "menu_text": "#000"
      }
    self.setup_build_directory_structure()
    self.get_git_info()
    self.main_menu_item_counter = 1
    self.specification_order = None
    self.specification_counter = 1
    self.specification_templates = []

  def setup_build_directory_structure(self):
    self.assets_directory_path = make_dir(os.path.join(self.build_directory, "assets"))
    self.content_directory_path = make_dir(os.path.join(self.build_directory, "content"))
    self.layouts_directory_path = make_dir(os.path.join(self.build_directory, "layouts"))
    self.static_directory_path = make_dir(os.path.join(self.build_directory, "static"))

    # Asset directories
    self.icon_directory_path = make_dir(os.path.join(self.assets_directory_path, "icons"))
    self.scss_directory_path = make_dir(os.path.join(self.assets_directory_path, "scss"))

    # Content directories
    self.about_directory_path = make_dir(os.path.join(self.content_directory_path, "about"))
    self.specifications_directory_path = make_dir(os.path.join(self.content_directory_path, "specifications"))
    self.schema_directory_path = make_dir(os.path.join(self.content_directory_path, "schema"))
    self.examples_directory_path = make_dir(os.path.join(self.content_directory_path, "examples"))

    # Copy layouts
    copy_tree(os.path.join(os.path.dirname(__file__),"hugo_layouts"),self.layouts_directory_path)

  def get_git_info(self):
    self.git_repo = git.Repo(self.docs_source_directory, search_parent_directories=True)
    self.git_remote_url = os.path.splitext(self.git_repo.remotes[0].url)[0]
    git_url_parts = self.git_remote_url.split('/')
    self.git_repo_name = git_url_parts[-1]
    self.git_repo_owner = git_url_parts[-2]
    self.git_repo_host = os.path.splitext(git_url_parts[-3])[0]
    self.git_branch_name = self.git_repo.active_branch.name
    self.base_url = fr"https://{self.git_repo_owner}.{self.git_repo_host}.io/{self.git_repo_name}/"

  def make_pages(self):
    # Check config directory
    landing_page_content = ""
    about_page_content = None
    background_image_path = None
    logo_path = None
    config_path = None
    favicons_path = None
    if os.path.exists(self.docs_config_directory):
      for file_name in os.listdir(self.docs_config_directory):
        file_path = os.path.join(self.docs_config_directory, file_name)
        if "landing" in file_name:
          with open(file_path,'r') as file:
            landing_page_content = file.read()
        elif "about" in file_name:
          with open(file_path,'r') as file:
            about_page_content = file.read()
        elif "featured-background" in file_name:
          background_image_path = file_path
        elif "logo" in file_name:
          logo_path = file_path
        elif "config" in file_name:
          config_path = file_path
        elif "favicons" in file_name:
          favicons_path = file_path

    if config_path is not None:
      config = load(config_path)
      if "title" in config:
        self.title = config["title"]
      if "author" in config:
        self.author = config["author"]
      if "description" in config:
        self.description = config["description"]
      if "specifications" in config:
        self.specification_order = config["specifications"]
      if "colors" in config:
        for item in config["colors"]:
          self.colors[item] = config["colors"][item]

    # Create scss file
    render_template(os.path.join(os.path.dirname(__file__),"scss_template.scss.j2"), os.path.join(self.scss_directory_path, "_variables_project.scss"), self.colors)

    landing_page_path = os.path.join(self.content_directory_path,"_index.pdc")
    self.make_page(landing_page_path,{"title": self.title, "type": "landing", "description": self.description}, content=landing_page_content)

    if about_page_content is not None:
      self.make_main_menu_page(self.about_directory_path,"About",content=about_page_content)

    if background_image_path is not None:
      shutil.copy(background_image_path, self.content_directory_path)
      if about_page_content is not None:
        shutil.copy(background_image_path, self.about_directory_path)

    if logo_path is not None:
      shutil.copy(logo_path, self.icon_directory_path)
      self.has_logo = True

    if favicons_path is not None:
      copy_tree(favicons_path, os.path.join(self.static_directory_path,"favicons"))

    # Specifications
    self.make_main_menu_page(self.specifications_directory_path,"Specifications",content="This data model contains the following specifications:",content_path=os.path.relpath(self.docs_source_directory),schema_path=os.path.relpath(self.source_schema_directory_path))

    self.make_specification_pages()

    # Schema
    self.make_main_menu_page(self.schema_directory_path,"Schema",content="Coming soon! Download the JSON Schema!")

    # Examples
    self.make_main_menu_page(self.examples_directory_path,"Examples",content="Coming soon! Download valid examples!")

  def make_specification_pages(self):
    # Collect list of doc template files
    if self.specification_order is not None:
      for specification_name in self.specification_order:
        file_name = f"{specification_name}.md.j2"
        file_path = os.path.join(self.docs_source_directory, file_name)
        if os.path.exists(file_path):
          self.specification_templates.append(DocumentFile(file_path))
        else:
          raise Exception(f"Unable to find specification file, \"{file_path}\", referenced in configuration.")
    else:
      for file_name in os.listdir(self.docs_source_directory):
        file_path = os.path.join(self.docs_source_directory, file_name)
        if os.path.isfile(file_path):
          self.specification_templates.append(DocumentFile(file_path))

    # Identify corresponding schema files
    for template in self.specification_templates:
      for schema_file in os.listdir(self.source_schema_directory_path):
        if template.file_base_name in schema_file:
          file_path = os.path.join(self.source_schema_directory_path, schema_file)
          template.set_corresponding_schema_path(file_path)


    # Process templates
    for template in self.specification_templates:
      template.set_markdown_output_path(os.path.join(self.specifications_directory_path,f"{get_file_basename(template.path, depth=2)}.pdc"))
      self.make_specification_page(template.path, template.markdown_output_path, self.source_schema_directory_path, template.corresponding_schema_path)

  def make_page(self, page_path, front_matter, content=""):
    with open(page_path,'w') as file:
      file.writelines(f"{make_front_matter(front_matter)}{content}")

  def make_main_menu_page(self, page_dir_path, title, content="", schema_path=None, content_type=None, content_path=None):
    front_matter = {
      "title": title,
      "linkTitle": title,
      "weight": 1,
      "menu": {
        "main": {
          "weight": self.main_menu_item_counter
        }
      }
    }

    self.main_menu_item_counter += 1

    if content_type is not None:
      front_matter["type"] = content_type

    if schema_path is not None:
      front_matter["github_schema"] = schema_path

    if content_path is not None:
      front_matter["github_content"] = content_path

    page_path = os.path.join(page_dir_path, "_index.pdc")
    self.make_page(page_path, front_matter, content)

  def make_specification_page(self, template_path, output_path, schema_dir_path, corresponding_schema_path=None):

    if get_extension(template_path) == ".j2":
      # Process template
      process_template(template_path, output_path, schema_dir=schema_dir_path)
    else:
      copy_tree(template_path, output_path)

    title = get_file_basename(template_path, depth=2)

    # Run Pandoc

    # Append front matter
    front_matter = {
      "title": title,
      "linkTitle": title,
      "type": "specifications",
      "weight": self.specification_counter
    }

    self.specification_counter += 1

    template_relative_path = os.path.relpath(template_path)

    if corresponding_schema_path is None:
      corresponding_schema_path = schema_dir_path

    front_matter["github_schema"] = corresponding_schema_path

    front_matter["github_content"] = template_relative_path

    with open(output_path,'r') as original_file:
      content = original_file.read()
    self.make_page(output_path, front_matter, content)

  def build(self):
    self.make_pages()

    # Setup Hugo Config
    dump(self.make_hugo_config(),os.path.join(self.build_directory,"config.yaml"))

    # npm package.json
    dump(self.make_npm_package_json(),os.path.join(self.build_directory,"package.json"))

    if not os.path.exists(os.path.join(self.build_directory,"go.mod")):
      subprocess.run(["hugo", "mod", "init", os.path.relpath(self.docs_source_directory)],cwd=self.build_directory,check=True)

    if not os.path.exists(os.path.join(self.build_directory,"go.sum")):
      subprocess.run(["hugo", "mod", "get", r"github.com/google/docsy@v0.4.0"],cwd=self.build_directory,check=True)

    if not os.path.exists(os.path.join(self.build_directory,"package-lock.json")):
      subprocess.run(["npm", "install"],cwd=self.build_directory,check=True)

    subprocess.run(["hugo", "--minify"],cwd=self.build_directory,check=True)

  def make_hugo_config(self):
    return {
      "baseURL": self.base_url,
      "title": self.title,
      "enableGitInfo": True,
      "module": {
        "hugoVersion": {
          "extended": True,
          "min": "0.73.0"
        },
        "imports": [
          {
            "path": "github.com/google/docsy",
            "disable": False
          },
          {
            "path": "github.com/google/docsy/dependencies",
            "disable": False
          },
        ]
      },
      "params": {
        "copyright": self.author,
        "github_repo": self.git_remote_url,
        "github_branch": self.git_branch_name,
        "ui": {
          "navbar_logo": self.has_logo,
          "breadcrumb_disable": True
        }
      },
      "security": {
        "enableInlineShortcodes": False,
        "exec": {
          "allow": ['^dart-sass-embedded$', '^go$', '^npx$', '^postcss$', '^pandoc$'],
          "osEnv": ['(?i)^(PATH|PATHEXT|APPDATA|TMP|TEMP|TERM)$']
        },
        "funcs": {
          "getenv": ['^HUGO_']
        },
        "http": {
          "methods": ['(?i)GET|POST'],
          "urls": ['.*']
        }
      }
    }

  def make_npm_package_json(self):
    return {
      "name": "lattice",
      "version": "0.0.1",
      "description": "",
      "dependencies": {},
      "devDependencies": {
        "autoprefixer": "^10.4.0",
        "postcss": "^8.3.7",
        "postcss-cli": "^9.0.2"
      }
    }

def make_front_matter(front_matter):
  return f"---\n{dump_to_string(front_matter,'yaml')}---\n\n"

def prepend_file_content(file_path, new_content):
  with open(file_path,'r') as original_file:
    original_content = original_file.read()
  with open(file_path,'w') as modified_file:
    modified_file.write(new_content + original_content)

def make_examples_html():
  pass

def render_template(template_path, output_path, values):
  template_directory_path = os.path.abspath(os.path.join(template_path, os.pardir))
  template_environment = Environment(loader=FileSystemLoader(template_directory_path))
  template = template_environment.get_template(os.path.basename(template_path))
  with open(output_path, 'w') as file:
    file.write(template.render(**values))



"""
Functionality to render Jinja templates with an add_schema_table hook to generate
schema tables in Markdown.
"""
import re
import os
import traceback

from jinja2 import Environment, FileSystemLoader, select_autoescape, TemplateNotFound
import yaml

from .schema_table import load_structure_from_object, data_types_table, string_types_table, enumerators_table, data_groups_table, write_data_model
from .grid_table import write_table

def make_args_string(args_dict):
    """
    - args_dict: dict, the dictionary of local variable to value such as returned by locals()
    RETURN: string, all arguments in a string
    """
    return ", ".join(
            [f"{k}={repr(v)}" for k, v in reversed(list(args_dict.items()))])


def log_error(error, log):
    """
    - error: string
    - log: None or list of string
    RETURN: error string
    """
    if log is not None:
        log.append(error)
    return error


def make_error_string(msg, args_str):
    """
    - msg: string, an error message
    - args_str: string, original arguments
    RETURN: string, an error message
    """
    return ("\n---\nERROR\n" + msg +
            f"\nin call to `add_schema_table({args_str})`\n---\n")


def render_header(level_and_content):
    """
    - level_and_content: None OR string OR list of [postive-int, string] OR tuple of
      (positive-int, string), the header level and content
    RETURN: string, the header
    """
    if level_and_content is None:
        return ""
    if isinstance(level_and_content, str):
        level = 1
        content = level_and_content
    elif isinstance(level_and_content, (list, tuple)):
        level, content = level_and_content
        if level < 1:
            level = 1
    else:
        raise Exception("Unhandle type of level_and_content")
    return "#" * level + " " + content + "\n\n"


def canonicalize_string(name):
    """
    Turns any string into a lowercase underscore single word
    - name: string, the table name
    RETURN: string, the name in canonical format
    """
    return "_".join([
        item.lower().strip()
        for item in re.split('\\s+', name.strip())])


def fetch_key_canonically(the_dict, canonical_key, default=None):
    """
    - the_dict: dictionary
    - canonical_key: any, the key of the dictionary
    - default: None or any, defaults to None. Returned if a canonical key match
      isn't found
    """
    for key, val in the_dict.items():
        if canonicalize_string(key) == canonical_key:
            return val
    return default


def extract_target_data(struct, table_name):
    """
    - struct: dict, raw data to pull from
    - table_name: string, assumed to be 'canonicalized' (via canonicalize_string)
    RETURN: Tuple of (None or string, None or dict or array, None or string for table type),
        - If there is an error, it is passed back as the first item; else None
        - If there is no error, the target data is passed back; else None if an error
        - If there is no error, we pass back the table type
    """
    if table_name in {'data_types', 'string_types'}:
        return (None, struct[table_name], table_name)
    for tbl in ['enumerations', 'data_groups']:
        table_type = tbl
        target = fetch_key_canonically(struct[tbl], table_name)
        if target is not None:
            break
    if target is None:
        return (
                f"`table_name` \"{table_name}\" was not `data_types` or `string_types`\n" +
                "and did not match any enumerators or data_groups in file!\n" +
                f"Possible enumerators: {', '.join(struct['enumerations'].keys())}\n" +
                f"Possible data_groups: {', '.join(struct['data_groups'].keys())}",
                None,
                None)
    return (None, target, table_type)


def determine_schema_dir(schema_dir):
    """
    - schema_dir: None or path-like, the path to the schema directory
    RETURN: path-like, the realized path
    """
    if schema_dir is None:
        # determine default path to package resources
        schema_dir = os.path.join(
                os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
                'schema-source')
    return schema_dir


def load_yaml_source(schema_dir, source, args_str):
    """
    - schema_dir: pathlike, the path to the schema directory
    - source: string, the source key. E.g., for
      schema-source/ASHRAE205.schema.yaml, 'ASHRAE205'
    - args_str: string, the arguments to the calling function (for error
      reporting)
    RETURN: (string or None, None or dict), tuple of the error string if didn't
            load or the data to return
    """
    src_path = os.path.join(schema_dir, source + '.schema.yaml')
    if not os.path.exists(src_path):
        return (
                make_error_string(
                    f"Schema source \"{source}\" (\"{src_path}\") doesn't exist!",
                    args_str),
                None)
    with open(src_path, encoding='utf-8', mode='r') as input_file:
        data = yaml.load(input_file, Loader=yaml.FullLoader)
    return (None, data)


def make_add_schema_table(schema_dir=None, error_log=None):
    """
    - schema_dir: string or pathlike, the path to the schema directory.
    - error_log: None or list, if a list, then errors will be appended to the
      log as well as rendered into the final product
    RETURN: returns the add_schema_table function with the following characteristics:
        - source: string, the source key. E.g., for
          schema-source/ASHRAE205.schema.yaml, 'ASHRAE205'
        - table_name: one of `data_types`, `string_types`, or a specific item
          from `enumerations`, or `data_groups`
        - caption: None or string, the table caption if desired
        - header_level_and_content: None OR Tuple of (positive-int, string), the
          header level and header content if desired
        RETURN: string, returns a string representation of the given table
    """
    schema_dir = determine_schema_dir(schema_dir)
    def add_schema_table(
            source,
            table_name,
            caption=None,
            header_level_and_content=None):
        args_str = make_args_string(locals())
        err, data = load_yaml_source(schema_dir, source, args_str)
        if err is not None:
            return log_error(err, error_log)
        struct = load_structure_from_object(data)
        err, target, table_type = extract_target_data(
                struct,
                canonicalize_string(table_name))
        if err is not None:
            return log_error(
                    make_error_string(err, args_str),
                    error_log)
        table_type_to_fn = {
                'data_types': data_types_table,
                'string_types': string_types_table,
                'enumerations': enumerators_table,
                'data_groups': data_groups_table,
                }
        gen_table = table_type_to_fn.get(table_type, None)
        if gen_table is None:
            return log_error(
                    make_error_string(
                        f"Unhandled table type \"{table_name}\"!",
                        args_str),
                    error_log)
        if caption is None:
            caption = table_name
        return render_header(header_level_and_content) + gen_table(
                target,
                caption=caption,
                add_training_ws=False)
    return add_schema_table

def make_add_yaml_table():
    """
    RETURN: returns the add_yaml_table function with the following characteristics:
        - content: string containing YAML syntax for Headers and Data.
        - caption: None or string, the table caption if desired
        RETURN: string, returns a string representation of the given table
    """
    def add_yaml_table(
            content,
            caption=None):
        table_data = yaml.load(content, Loader=yaml.FullLoader)
        columns = table_data["Headers"]
        data = {}
        for i, column in enumerate(columns):
            data[column] = []
            for row in table_data["Data"]:
                data[column].append(row[i])
        return write_table(data, columns, caption)
    return add_yaml_table


def make_add_data_model(schema_dir, error_log):
    """
    - schema_dir: string or pathlike, the path to the schema directory.
    - error_log: None or list, if a list, then errors will be appended to the
      log as well as rendered into the final product
    RETURN: returns the add_data_model function with the following characteristics:
        - source: string, the source key. E.g., for
          schema-source/ASHRAE205.schema.yaml, 'ASHRAE205'
        RETURN: string, returns a string representation of the given data models
    """
    schema_dir = determine_schema_dir(schema_dir)
    def add_data_model(source, make_headers=True, base_level=1):
        args_str = make_args_string(locals())
        err, data = load_yaml_source(schema_dir, source, args_str)
        if err is not None:
            return log_error(err, error_log)
        return write_data_model(data, make_headers, base_level)
    return add_data_model


def process_template(template_path, output_path, schema_dir=None, log_file=None):
    """
    - template_path: string, path to the main template file.
    - output_path: string, the output path to write the template to
    - schema_dir: string, a custom path to the schema files (*.schema.yaml) to work from
    - log_file: string or None, if a string, the path to an error file to write
    RETURN: None
    SIDE-EFFECTS:
    - load the template main_template from template_dir
    - render that template
    - write the contents to output_path
    """
    template_dir = os.path.abspath(os.path.join(template_path, os.pardir))
    env = Environment(
        loader=FileSystemLoader(template_dir),
        autoescape=select_autoescape(['html', 'xml']),
        trim_blocks=True,
        lstrip_blocks=True,
        comment_start_string="{##",
        comment_end_string="##}")
    errs = None # errors
    if log_file is not None:
        errs = []
    try:
        temp = env.get_template(os.path.basename(template_path))
        with open(output_path, encoding='utf-8', mode='w') as handle:
            handle.write(
                    temp.render(
                        add_schema_table=make_add_schema_table(schema_dir, errs),
                        add_yaml_table=make_add_yaml_table(),
                        add_data_model=make_add_data_model(schema_dir, errs)))
    except TemplateNotFound as exc:
        print(f"Could not find main template {template_path}")
        print(f"template_path = {template_path}")
        print(f"output_path = {output_path}")
        print(f"template_dir = {template_dir}")
        print("Exception:")
        print(exc)
        traceback.print_exc()
    if log_file is not None:
        with open(log_file, 'w') as handle:
            if len(errs) > 0:
                for err in errs:
                    handle.write(err.strip())
                    handle.write("\n")



import os
import json
import posixpath
import jsonschema
import yaml
import sys
import re
import copy

from .file_io import load, dump, get_file_basename
from .schema_to_json import JSON_translator


class MetaSchema:
  def __init__(self, schema_path):
    with open(schema_path) as meta_schema_file:
      uri_path = os.path.abspath(os.path.dirname(schema_path))
      if os.sep != posixpath.sep:
        uri_path = posixpath.sep + uri_path

      resolver = jsonschema.RefResolver(f'file://{uri_path}/', meta_schema_file)
      self.validator = jsonschema.Draft7Validator(json.load(meta_schema_file), resolver=resolver)

  def validate(self, instance_path):
    with open(os.path.join(instance_path), 'r') as input_file:
      instance = yaml.load(input_file, Loader=yaml.FullLoader)
    errors = sorted(self.validator.iter_errors(instance), key=lambda e: e.path)
    file_name =  os.path.basename(instance_path)
    if len(errors) == 0:
      print(f"Validation successful for {file_name}")
    else:
      messages = []
      for error in errors:
        messages.append(f"{error.message} ({'.'.join([str(x) for x in error.path])})")
      messages = [f"{i}. {message}" for i, message in enumerate(messages, start=1)]
      message_str = '\n  '.join(messages)
      raise Exception(f"Validation failed for {file_name} with {len(messages)} errors:\n  {message_str}")

def meta_validate_file(file_path, meta_schema_path):
  meta_schema = MetaSchema(meta_schema_path)
  meta_schema.validate(file_path)

# Generate Meta Schema

class SchemaTypes:

    def __init__(self, core_schema, schema=None):
        # Generate Regular Expressions

        # Data Types
        core_types = get_types(core_schema)
        self.combined_types = set(core_types.keys())

        regex_base_types = core_types["Data Type"]

        base_types = '|'.join(regex_base_types)
        base_types = f"({base_types})"

        string_types = core_types["String Type"]
        if schema:
            schema_types = get_types(schema)
            self.combined_types |= set(schema_types.keys())
            if "String Type" in schema_types:
                string_types += ["String Type"]

        re_string_types = '|'.join(string_types)
        re_string_types = f"({re_string_types})"

        template_data_group_prefixes = []
        if schema:
            template_data_group_prefixes = [key for key in schema if schema[key]['Object Type'] == 'Data Group Template']

        self.type_base_names = "[A-Z]([A-Z]|[a-z]|[0-9])*"
        self.type_base_names_anchored = f"^{self.type_base_names}$"
        if len(template_data_group_prefixes) > 0:
            self.data_group_names_anchored = f"(^(?!({'|'.join(template_data_group_prefixes)})){self.type_base_names}$)"
        else:
            self.data_group_names_anchored = self.type_base_names_anchored
        data_groups = fr"\{{{self.type_base_names}\}}"
        enumerations = fr"<{self.type_base_names}>"
        optional_base_types = fr"{base_types}{{1}}(\/{base_types})*"
        single_type = fr"({optional_base_types}|{re_string_types}|{data_groups}|{enumerations})"
        alternatives = fr"\(({single_type})(,\s*{single_type})+\)"
        arrays = fr"\[({single_type})\](\[\d*\.*\d*\])?"
        data_types = f"({single_type})|({alternatives})|({arrays})"
        self.data_types_anchored = f"^{data_types}$"

        pattern = re.compile(self.data_types_anchored)

        valid_tests = [
            "Numeric",
            "[Numeric]",
            "{DataGroup}",
            "[String][1..]",
            ]

        invalid_tests = [
            "Wrong"
        ]

        for test in valid_tests:
            if not re.search(pattern,test):
                raise Exception(f"\"{test}\" does not match: {pattern}")

        for test in invalid_tests:
            if re.search(pattern,test):
                raise Exception(f"\"{test}\" matches: {pattern}")

        # Values
        number = "([-+]?[0-9]*\\.?[0-9]+([eE][-+]?[0-9]+)?)"
        string = "\".*\""
        enumerator = "([A-Z]([A-Z]|[0-9])*)(_([A-Z]|[0-9])+)*"
        boolean = "True|False"
        values = f"({number})|({string})|({enumerator})|({boolean})"

        self.enumerator_anchored = f"^{enumerator}$"

        re.compile(values)

        # Constraints
        self.element_names = "([a-z]+)(_([a-z]|[0-9])+)*"
        self.element_names_anchored = f"^{self.element_names}$"
        alpha_array = "(\[A-Z\]{[1-9]+})"
        numeric_array = "(\[0-9\]{[1-9]+})"
        ranges = f"(>|>=|<=|<){number}"
        multiples = f"%{number}"
        data_element_value = f"{self.element_names}={values}"
        sets = f"\[{number}(, {number})*\]"
        reference_scope = f":{self.type_base_names}:"
        selector = fr"{self.element_names}\({values}(,\s*{values})*\)"

        constraints = f"({alpha_array}|{numeric_array}|{ranges})|({multiples})|({sets})|({data_element_value})|({reference_scope})|({selector})"
        self.constraints_anchored = f"^{constraints}$"

        re.compile(constraints)

        # Conditional Requirements
        conditional_requirements = f"if (!?{self.element_names})(!?=({values}))?"
        self.conditional_requirements_anchored = f"^{conditional_requirements}$"

        re.compile(conditional_requirements)


core_meta_schema_path = os.path.join(os.path.dirname(__file__),'meta.schema.yaml')
core_schema_path = os.path.join(os.path.dirname(__file__),'core.schema.yaml')

def generate_meta_schema(output_path, schema=None):
    ''' '''
    core_schema = load(core_schema_path)

    referenced_schema = []
    if schema is not None:
        source_schema = load(schema)
        schema_dir = os.path.dirname(schema)
        if 'References' in source_schema:
            referenced_schema = [load(schema_file) for schema_file in [os.path.join(schema_dir, f'{ref}.yaml') for ref in source_schema['References']]]
        meta_schema_file_name = f"{get_file_basename(schema, depth=2)}.meta.schema.json"
    else:
        source_schema = None

    schematypes = SchemaTypes(core_schema, source_schema)

    meta_schema = load(core_meta_schema_path)

    # Replace generated patterns from core schema analysis

    meta_schema["definitions"]["Meta"]["properties"]["Unit Systems"]["patternProperties"][schematypes.type_base_names_anchored] = meta_schema["definitions"]["Meta"]["properties"]["Unit Systems"]["patternProperties"].pop("**GENERATED**")
    #meta_schema["definitions"]["DataGroupTemplate"]["patternProperties"][schematypes.type_base_names_anchored] = meta_schema["definitions"]["DataGroupTemplate"]["patternProperties"].pop("**GENERATED**")
    meta_schema["definitions"]["DataGroupTemplate"]["properties"]["Required Data Elements"]["patternProperties"][schematypes.element_names_anchored] = meta_schema["definitions"]["DataGroupTemplate"]["properties"]["Required Data Elements"]["patternProperties"].pop("**GENERATED**")
    meta_schema["definitions"]["DataGroupTemplate"]["properties"]["Object Type"]["pattern"] = schematypes.type_base_names
    meta_schema["definitions"]["ConstraintsPattern"]["pattern"] = schematypes.constraints_anchored
    meta_schema["definitions"]["Required"]["oneOf"][1]["pattern"] = schematypes.conditional_requirements_anchored
    meta_schema["definitions"]["Enumerator"]["patternProperties"][schematypes.enumerator_anchored] = meta_schema["definitions"]["Enumerator"]["patternProperties"].pop("**GENERATED**")
    meta_schema["definitions"]["DataTypePattern"]["pattern"] = schematypes.data_types_anchored
    meta_schema["definitions"]["DataGroup"]["properties"]["Data Elements"]["patternProperties"][schematypes.element_names_anchored] = meta_schema["definitions"]["DataGroup"]["properties"]["Data Elements"]["patternProperties"].pop("**GENERATED**")
    meta_schema["patternProperties"][schematypes.data_group_names_anchored] = meta_schema["patternProperties"].pop("**GENERATED**")

    # Replace generated patterns from schema instance

    if source_schema:
        if 'Schema' in source_schema:
            if 'Title' in source_schema['Schema']:
                meta_schema["title"] = f"{source_schema['Schema']['Title']} Meta-schema"
            # Special Unit Systems
            if 'Unit Systems' in source_schema['Schema']:
                # Unit system definitions
                for unit_system in source_schema['Schema']["Unit Systems"]:
                    meta_schema["definitions"][unit_system] = {"type": "string", "enum": source_schema['Schema']["Unit Systems"][unit_system]}
                    meta_schema["definitions"]["DataElementAttributes"]["properties"]["Units"]["anyOf"].append({"$ref":f"meta.schema.json#/definitions/{unit_system}"})
        # Special Data Groups (gather from referenced Schema as well)
        combined_schema = source_schema.copy()
        for d in referenced_schema:
            combined_schema.update(d)
        for data_group_type in [key for key in combined_schema if combined_schema[key]['Object Type'] == 'Data Group Template']:
            # Data Element Attributes
            data_group = source_schema[data_group_type]
            data_group_type_name = data_group["Name"]
         
            for template_data_group in [template_key for template_key in source_schema if data_group_type_name == combined_schema[template_key]['Object Type']]:
                meta_schema["definitions"][f"{template_data_group}DataElementAttributes"] = copy.deepcopy(meta_schema["definitions"]["DataElementAttributes"])

                if "Unit System" in data_group:
                    meta_schema["definitions"][f"{template_data_group}DataElementAttributes"]["properties"]["Units"] = {"$ref": f"meta.schema.json#/definitions/{data_group['Unit System']}"}

                if "Required Data Types" in data_group:
                    meta_schema["definitions"][f"{template_data_group}DataElementAttributes"]["properties"]["Data Type"] = {"type": "string", "enum": data_group["Required Data Types"]}

                if "Data Elements Required" in data_group:
                    meta_schema["definitions"][f"{template_data_group}DataElementAttributes"]["properties"]["Required"]["const"] = data_group["Data Elements Required"]

                # Data Group
                meta_schema["definitions"][f"{template_data_group}DataGroup"] = copy.deepcopy(meta_schema["definitions"]["DataGroup"])
                meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Object Type"]["const"] = data_group["Name"]

                if "Required Data Elements" in data_group:
                    required_names = []
                    meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["properties"] = {}
                    for data_element_name in data_group["Required Data Elements"]:
                        required_names.append(data_element_name)
                        data_element = data_group["Required Data Elements"][data_element_name]
                        meta_schema["definitions"][f"{template_data_group}-{data_element_name}-DataElementAttributes"] = copy.deepcopy(meta_schema["definitions"][f"{template_data_group}DataElementAttributes"])
                        meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["properties"][data_element_name] = copy.deepcopy(meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["patternProperties"][schematypes.element_names_anchored])
                        meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["properties"][data_element_name]["$ref"] = f"meta.schema.json#/definitions/{template_data_group}-{data_element_name}-DataElementAttributes"
                        for attribute in data_element:
                            meta_schema["definitions"][f"{template_data_group}-{data_element_name}-DataElementAttributes"]["properties"][attribute]["const"] = data_element[attribute]
                            if attribute not in meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["properties"][data_element_name]["required"]:
                                meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["properties"][data_element_name]["required"].append(attribute)
                    meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["required"] = required_names
                    exclusive_element_names_anchored = f"(?!^({'|'.join(required_names)})$)(^{schematypes.element_names}$)"
                    meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["patternProperties"][exclusive_element_names_anchored] = meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["patternProperties"].pop(schematypes.element_names_anchored)
                    meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["patternProperties"][exclusive_element_names_anchored]["$ref"] = f"meta.schema.json#/definitions/{template_data_group}DataElementAttributes"
                else:
                    meta_schema["definitions"][f"{template_data_group}DataGroup"]["properties"]["Data Elements"]["patternProperties"][schematypes.element_names_anchored]["$ref"] = f"meta.schema.json#/definitions/{template_data_group}DataElementAttributes"

                # Main schema
                meta_schema["patternProperties"][f"(^{template_data_group}*$)"] = {"$ref": f"meta.schema.json#/definitions/{template_data_group}DataGroup"}
            meta_schema["patternProperties"][f"^{data_group_type}$"] = {"$ref": f"meta.schema.json#/definitions/DataGroupTemplate"}

    dump(meta_schema, output_path)

    with open(output_path, 'r') as file:
      content = file.read()
    with open(output_path, 'w') as file:
      file.writelines(content.replace("meta.schema.json", meta_schema_file_name))

    return schematypes


def get_types(schema):
    '''For each Object Type in a schema, map a list of Objects matching that type.'''
    types = {}
    for object in schema:
        if schema[object]["Object Type"] not in types:
            types[schema[object]["Object Type"]] = []
        types[schema[object]["Object Type"]].append(object)
    return types



"""
Markdown grid-table creation utilities
"""
import copy
import io
import stringcase

def flatten(list_of_lists):
    """
    - list_of_lists: (list (list *)), a list of lists
    RETURN: (List *), the list flattened
    """
    return [
        item
        for sublist in list_of_lists
        for item in sublist]


def wrap_text_to_lines(text, width, bold=False, left_space=1, right_space=1):
    """
    - text: string, the text to wrap
    - width: int, >0, the width to wrap to
    - bold: bool, if True, allow space for surrounding the beginning and end with **
    - left_space: int, >=0, the left space to enforce between cells
    - right_space: int, >=0, the right space to enforce between cells
    NOTE: width includes both left_space and right_space
    RETURN: (Array String), the lines
    """
    verbose = False
    if len(text) > 0:
        if text[0:1] in ['+ ','- ','* ']:
            text = '\\' + text
    if "\n" in text:
        return flatten([
            wrap_text_to_lines(t, width, bold, left_space, right_space)
            for t in text.split("\n")])
    atoms = text.split(" ")
    if bold:
        atoms[0] = '**' + atoms[0]
        atoms[-1] = atoms[-1] + '**'
    longest_atom = max([len(s) for s in atoms])
    if longest_atom + left_space + right_space > width:
        print("Warning! Need to hyphenate atoms!")
    lines = []
    atom_idx = 0
    first = True
    line = ' '*left_space
    num_atoms = len(atoms)
    while atom_idx < num_atoms:
        if verbose:
            print(f"atom_idx: {atom_idx}")
            print(f"line    : {line}")
            print(f"lines   : {lines}")
            print(f"atoms   : {atoms}")
        if first:
            new_line = line + atoms[atom_idx]
            first = False
        else:
            new_line = line + ' ' + atoms[atom_idx]
        if len(new_line) + right_space <= width:
            line = new_line
        else:
            lines.append(line + ' '*right_space)
            line = (' '*left_space) + atoms[atom_idx]
        atom_idx += 1
    if len(line) > 0:
        lines.append(line + ' '*right_space)
    if verbose:
        print(f"line    : {line}")
        print(f"lines   : {lines}")
        print(f"atoms   : {atoms}")
    return lines


def get_line_row(lines, row):
    """
    - lines: (Array string), array of lines
    - row: int, >=0, the row index to grab
    RETURN: string, if row greater than or equal to lines length, returns ''
    """
    if row < len(lines):
        return lines[row]
    return ''


def write_rule(sizes, is_header=False):
    """
    - sizes: (Array int), the column widths (of contents)
    RETURN: string
    """
    mark = '=' if is_header else '-'
    return '+' + ''.join([(mark*n + '+') for n in sizes]) + '\n'


def write_row(content, sizes, is_header=False):
    """
    RETURN: string
    """
    assert len(content) == len(sizes), 'len(content) must equal len(sizes)'
    table = write_rule(sizes) if is_header else ''
    list_of_lines = [
            wrap_text_to_lines(text=content[n], width=sizes[n], bold=is_header)
            for n in range(len(content))]
    for row_num in range(max([len(lines) for lines in list_of_lines])):
        table_line = '|' + ''.join(['{:' + str(n)  + 's}|' for n in sizes]) + '\n'
        texts = [get_line_row(lines, row_num) for lines in list_of_lines]
        table += table_line.format(*texts)
    table += write_rule(sizes, is_header=is_header)
    return table


def get_column_sizes(content, is_bold=False, has_spacing=True, preferred_sizes=None, full_width=100):
    """
    - content: (Array string), the column content
    - is_bold: Bool, True if the column content must be surrounded with '**'
    - has_spacing: Bool, True if we keep at least one space padding between cell border
    - preferred_sizes: None or (Array int), the preferred sizes of the columns.
      Will be at least this size
    - full_width: Should be same as --columns Pandoc argument
      Pandoc default=72
    RETURN: (Array int), the actual sizes
    """
    if preferred_sizes is None:
        sizes = [0] * len(content)
    else:
        sizes = copy.deepcopy(preferred_sizes)
    full_sizes = copy.deepcopy(sizes)
    size_diff = [0] * len(sizes)
    max_diff_col = 0
    assert len(content) == len(sizes), "len(content) must equal len(sizes)"
    for col_num, col_content in enumerate(content):
        full_size = len(col_content)
        atoms = col_content.replace('\n',' ').split(' ')
        longest_atom = max([len(s) for s in atoms])
        if is_bold:
            num_atoms = len(atoms)
            if num_atoms == 1:
                longest_atom += 4
            elif num_atoms > 1:
                longest_atom += 2
            full_size += 4
        if has_spacing:
            longest_atom += 2
            full_size += 2
        if longest_atom > sizes[col_num]:
            sizes[col_num] = longest_atom
        if full_size > full_sizes[col_num]:
            full_sizes[col_num] = full_size
        size_diff[col_num] = full_sizes[col_num] - sizes[col_num]
        if size_diff[col_num] > size_diff[max_diff_col]:
            max_diff_col = col_num
    if sum(full_sizes) < full_width:
        # full text will fit on one line, no need to make anything smaller
        return full_sizes
    elif sum(sizes) >= full_width:
        # Sizes are too big already...things will be too tight to make any changes
        return sizes
    else:
        # Relax some sizes to fill in full width
        # Weigh towards columns with more text
        remaining_space = full_width - sum(sizes)
        total_diff = sum(size_diff)
        for col_num in range(len(sizes)):
            sizes[col_num] += size_diff[col_num]*remaining_space//total_diff
        # Check remaining size
        remaining_space = full_width - sum(sizes)
        if remaining_space > 0:
            # Add remaining space to column with largest size diff
            sizes[max_diff_col] += remaining_space

        return sizes

def check_dict_of_arrays(doa, columns):
    """
    Checks the data-structure that dict of arrays has at least the keys in
    columns and that each entry's length is the same as the others.
    - doa: (Dict String (Array String)), dictionary with string keys to arrays of string
    - columns: (Array String), the columns of doa
    RETURN: (Array String), array of issues
    """
    issues = []
    length = None
    for col in columns:
        if col not in doa:
            issues.append(f"{col} not in doa")
            continue
        if length is None:
            try:
                length = len(doa[col])
            except TypeError:
                issues.append(f"could not take length of doa['{col}']")
        elif len(doa[col]) != length:
            issues.append(
                    f"len(doa['{col}']) = {len(doa[col])} != {length}," +
                    " the length of other columns")
    return issues


def assert_doa_valid(doa, columns):
    """
    Checks if DOA valid. If not, raises an exception.
    - doa: (Dict String (Array String)), dictionary with string keys to arrays of string
    - columns: (Array String), order of keys to write table out as
    """
    issues = check_dict_of_arrays(doa, columns)
    if len(issues) > 0:
        raise Exception("DictOfArraysInvalid: " + ";".join(issues))


def remove_blank_columns(doa, columns, sizes):
    """
    Remove columns that are blank.
    - doa: (Dict String (Array String)), dictionary with string keys to arrays of string
    - columns: (Array String), order of keys to write table out as
    RETURN: (Tuple (Array String), (Array Integer)), tuple of new columns with
    blanks removed and new sizes
    """
    new_columns = []
    new_sizes = []
    for col, size in zip(columns, sizes):
        non_blanks = 0
        for row in doa[col]:
            if row is not None and row != "":
                non_blanks += 1
        if non_blanks > 0:
            new_columns.append(col)
            new_sizes.append(size)
    return new_columns, new_sizes


def make_table_from_dict_of_arrays(doa, columns, preferred_sizes=None, drop_blank_columns=False):
    """
    - doa: (Dict String (Array String)), dictionary with string keys to arrays of string
    - columns: (Array String), order of keys to write table out as
    - preferred_sizes: None or (Array Int) with preferred column sizes
    - drop_blank_columns: Bool, if True, drops columns where there are no
      entries for any rows in that columns
    RETURN: String, a grid table
    """
    if preferred_sizes is None:
        preferred_sizes = [0] * len(columns)
    if drop_blank_columns:
        columns, preferred_sizes = remove_blank_columns(doa, columns, preferred_sizes)
    assert_doa_valid(doa, columns)
    full_width = 100
    sizes = get_column_sizes(
            columns, is_bold=True, has_spacing=True,
            preferred_sizes=preferred_sizes,full_width=full_width)
    num_rows = len(doa[columns[0]])
    rows = []
    for row_num in range(num_rows):
        row = [doa[c][row_num] for c in columns]
        rows.append(row)
        sizes = get_column_sizes(
                row, is_bold=False, has_spacing=True, preferred_sizes=sizes, full_width=full_width)
    table = write_row(columns, sizes, True)
    for row in rows:
        table += write_row(row, sizes, False)
    return table


def write_table(dat, columns, caption, preferred_sizes=None):
    """
    - dat: (Dict String (Array String)), dict of arrays of data for the table
    - columns: (Array String), the column names in desired order
    - path: string, path to where to save the table
    - caption: None or string
    - preferred_sizes: None or (Array Integer), the preferred column sizes;
      column will be at least that size
    RETURN: string of the table in Markdown
    """
    if preferred_sizes is None:
        preferred_sizes = [0] * len(columns)
    the_str = ""
    with io.StringIO() as handle:
        handle.write(make_table_from_dict_of_arrays(
            dat, columns=columns, preferred_sizes=preferred_sizes))
        if caption is not None:
            handle.write(f"\nTable: {caption} {{#tbl:{stringcase.snakecase(caption)}}}\n")
        the_str = handle.getvalue()
    return the_str


def write_out_table(dat, columns, path, caption, preferred_sizes=None):
    """
    - dat: (Dict String (Array String)), dict of arrays of data for the table
    - columns: (Array String), the column names in desired order
    - path: string, path to where to save the table
    - caption: None or string
    - preferred_sizes: None or (Array Integer), the preferred column sizes;
      column will be at least that size
    - table_size: None or string, if string, one of "Huge", "huge", "LARGE",
      "Large", "large", "normalsize", "small", "footnotesize", "scriptsize",
      "tiny" the table size
    RETURN: None
    SIDE_EFFECT: create table and write it to path as a pandoc grid table
    """
    with open(path, 'w', encoding='utf-8') as handle:
        handle.write(
                write_table(
                    dat, columns, caption, preferred_sizes))



import os
import re
from fnmatch import fnmatch
import warnings

from lattice.docs.process_template import process_template
from .file_io import check_dir, make_dir, load, get_file_basename
from .meta_schema import generate_meta_schema, meta_validate_file
from .schema_to_json import generate_json_schema, validate_file, postvalidate_file
from .docs import HugoWeb, DocumentFile

class SchemaFile:
  def __init__(self, path):
    self.path = os.path.abspath(path)
    self.content = load(self.path)

    self.file_base_name = get_file_basename(self.path, depth=2)

    # Check for required content
    if "Schema" not in self.content:
      raise Exception(f"Required \"Schema\" object not found in schema file, \"{self.path}\".")

    self.schema_type = self.file_base_name # Overwritten if it is actually specified
    self.data_model_type = None
    self.root_data_group = None
    if "Root Data Group" in self.content["Schema"]:
      self.root_data_group = self.content["Schema"]["Root Data Group"]
      self.schema_type = self.root_data_group
      if self.root_data_group in self.content:
        # Get metadata
        self.data_model_type = None
        if "Data Elements" not in self.content[self.root_data_group]:
          raise Exception(f"Root Data Group, \"{self.root_data_group}\" does not contain \"Data Elements\".")
        else:
          if "metadata" in self.content[self.root_data_group]["Data Elements"]:
            if "Constraints" in self.content[self.root_data_group]["Data Elements"]["metadata"]:
              constraints = self.content[self.root_data_group]["Data Elements"]["metadata"]["Constraints"]
              data_element_pattern = "([a-z]+)(_([a-z]|[0-9])+)*"
              enumeration_pattern = "[A-Z]([A-Z]|[a-z]|[0-9])*"
              constraint_pattern = re.compile(f"^({data_element_pattern})=({enumeration_pattern})$")
              for constraint in constraints:
                match = constraint_pattern.match(constraint)
                if match:

                  if match.group(1) == "data_model":
                    self.data_model_type = match.group(5)
                  else:
                    pass # Warning?

                  if match.group(1) == "schema":
                    self.schema_type = match.group(5)
                  else:
                    pass # Warning?

          else:
            pass # Warning?

      else:
        raise Exception(f"Root Data Group, \"{self.root_data_group}\", not found in schema file, \"{self.path}\"")


      # TODO: Version?



  def set_meta_schema_path(self, meta_schema_path):
    self.meta_schema_path = os.path.abspath(meta_schema_path)

  def set_json_schema_path(self, json_schema_path):
    self.json_schema_path = os.path.abspath(json_schema_path)

  def set_schema_patterns(self, patterns):
    self.schema_patterns = patterns


class Lattice:
  def __init__(self, root_directory=".", build_directory=None, build_output_directory_name=".lattice/", build_validation=True):
    # Check if directories exists
    check_dir(root_directory, "Root directory")
    if build_directory is None:
      build_directory = root_directory
    else:
      check_dir(build_directory, "Build directory")

    self.root_directory = root_directory

    if build_output_directory_name is None:
      self.build_directory = build_directory
    else:
      self.build_directory = os.path.join(build_directory, build_output_directory_name)
    make_dir(self.build_directory)

    self.collect_schemas()

    self.setup_meta_schemas()

    self.setup_json_schemas()

    self.collect_example_files()

    self.collect_doc_templates()

    self.web_docs_directory_path = os.path.join(self.doc_output_dir,"web")

    if build_validation:
      self.generate_meta_schemas()
      self.validate_schemas()
      self.generate_json_schemas()
      self.validate_example_files()


  def collect_schemas(self):
    # Make sure root directory has minimum content (i.e., schema directory, or at least one schema file)
    schema_directory_path = os.path.join(self.root_directory,"schema")
    if os.path.exists(schema_directory_path):
      self.schema_directory_path = schema_directory_path
    else:
      self.schema_directory_path = self.root_directory

    # Collect list of schema files
    self.schemas = []
    for file_name in os.listdir(self.schema_directory_path):
      file_path = os.path.join(self.schema_directory_path, file_name)
      if fnmatch(file_path, "*.schema.yaml") or fnmatch(file_path, "*.schema.yml"):
        self.schemas.append(SchemaFile(file_path))

    if len(self.schemas) == 0:
      raise Exception(f"No schemas found in \"{self.schema_directory_path}\".")

  def setup_meta_schemas(self):
    self.meta_schema_directory = os.path.join(self.build_directory,"meta_schema")
    make_dir(self.meta_schema_directory)
    for schema in self.schemas:
      meta_schema_path = os.path.join(self.meta_schema_directory,f"{schema.file_base_name}.meta.schema.json")
      schema.set_meta_schema_path(meta_schema_path)

  def generate_meta_schemas(self):
    for schema in self.schemas:
      schema.set_schema_patterns(generate_meta_schema(schema.meta_schema_path, schema.path))

  def validate_schemas(self):
    for schema in self.schemas:
      meta_validate_file(schema.path, schema.meta_schema_path)

  def setup_json_schemas(self):
    self.json_schema_directory = os.path.join(self.build_directory,"json_schema")
    make_dir(self.json_schema_directory)
    self.core_json_schema_path = os.path.join(self.json_schema_directory,f"core.schema.json")
    for schema in self.schemas:
      json_schema_path = os.path.join(self.json_schema_directory,f"{schema.file_base_name}.schema.json")
      schema.set_json_schema_path(json_schema_path)

  def generate_json_schemas(self):
    for schema in self.schemas:
      generate_json_schema(schema.path, schema.json_schema_path)

  def validate_file(self, input_path, schema_type=None):
    instance = load(input_path)
    if schema_type is None:
      if "metadata" in instance:
        if "schema" in instance["metadata"]:
          schema_type = instance["metadata"]["schema"]

    if schema_type is None:
      if len(self.schemas) > 1:
        raise Exception(f"Multiple schemas defined, and no schema type provide. Unable to validate file, \"{input_path}\".")
      else:
        validate_file(input_path, self.schemas[0].json_schema_path)
        postvalidate_file(input_path, self.schemas[0].json_schema_path)
    else:
      # Find corresponding schema
      for schema in self.schemas:
        if schema.schema_type == schema_type:
          validate_file(input_path, schema.json_schema_path)
          postvalidate_file(input_path, schema.json_schema_path)
          break

  def collect_example_files(self):
    example_directory_path = os.path.join(self.root_directory,"examples")
    if os.path.exists(example_directory_path):
      self.example_directory_path = example_directory_path
    else:
      self.example_directory_path = None

    # Collect list of example files
    self.examples = []
    if self.example_directory_path is not None:
      for file_name in os.listdir(self.example_directory_path):
        file_path = os.path.join(self.example_directory_path, file_name)
        if os.path.isfile(file_path):
          self.examples.append(os.path.abspath(file_path))

  def validate_example_files(self):
    for example in self.examples:
      self.validate_file(example)

  def collect_doc_templates(self):
    doc_templates_directory_path = os.path.join(self.root_directory,"docs")
    if os.path.exists(doc_templates_directory_path):
      self.doc_templates_directory_path = doc_templates_directory_path
    else:
      self.doc_templates_directory_path = None

    # Collect list of doc template files
    self.doc_templates = []
    if self.doc_templates_directory_path is not None:
      for file_name in os.listdir(self.doc_templates_directory_path):
        file_path = os.path.join(self.doc_templates_directory_path, file_name)
        if os.path.isfile(file_path) and ".md" in file_name:
          self.doc_templates.append(DocumentFile(file_path))
    self.doc_output_dir = os.path.join(self.build_directory,"docs")
    if len(self.doc_templates) > 0:
      make_dir(self.doc_output_dir)
      for template in self.doc_templates:
        markdown_path = os.path.join(self.doc_output_dir,f"{get_file_basename(template.path, depth=1)}")
        template.set_markdown_output_path(markdown_path)

  def generate_markdown_documents(self):
    for template in self.doc_templates:
      process_template(template.path, template.markdown_output_path, self.schema_directory_path)

  def generate_web_documentation(self):
    make_dir(self.doc_output_dir)
    make_dir(self.web_docs_directory_path)

    if self.doc_templates_directory_path:
        HugoWeb(self.doc_templates_directory_path, self.web_docs_directory_path).build()
    else:
        warnings.warn('Template directory "doc" does not exist under {self.root_directory}')




"""
Specifics for setting up schema tables.
"""
from copy import deepcopy
import io
import re

from .grid_table import write_table

def write_header(heading, level=1):
    """
    - heading: string, the heading
    - level: integer, level > 0, the markdown level
    RETURN: string
    """
    return ("#"*level) + " " + heading + "\n\n"


def process_string_types(string_types):
    """
    - string_types: array of dict, the string types
    RETURN: list of dict, copy of string types list with regexes handled
    properly
    """
    new_list = []
    for str_typ in string_types:
        new_item = deepcopy(str_typ)
        if 'Is Regex' in new_item and new_item['Is Regex']:
            new_item['JSON Schema Pattern'] = '(Not applicable)'
        new_item['JSON Schema Pattern'] = new_item['JSON Schema Pattern'].replace('*',r'\*').replace(r'(?','\n'r'(?').replace(r'-[','\n'r'-[')
        new_list.append(new_item)
    return new_list


def compress_notes(a_dict):
    """
    - a_dict: Dict, a dictionary that may contain the key, 'Notes"
    RETURN:
    None
    SIDE-EFFECTS:
    modifies d in place to replace the "Notes" value with a string if it is an
    array.
    """
    notes = "Notes"
    if notes in a_dict:
        if isinstance(a_dict[notes], list):
            a_dict[notes] = "\n\n".join([f"- {note}" for note in a_dict[notes]])


def data_elements_dict_from_data_groups(data_groups):
    """
    - data_groups: Dict, the data groups dictionary
    RETURN: Dict with data elements as an array
    """
    output = {}
    for dat_gr in data_groups:
        data_elements = []
        for element in data_groups[dat_gr]["Data Elements"]:
            new_obj = data_groups[dat_gr]["Data Elements"][element]
            new_obj["Name"] = f"`{element}`"
            if 'Required' in new_obj:
                if new_obj["Required"] == True:
                    check = u'\N{check mark}'
                    new_obj["Req"] = f"${check}$" if new_obj["Required"] else ''
                elif new_obj["Required"] == False:
                    new_obj["Req"] = ''
                else:
                    new_obj["Req"] = f"`{new_obj['Required']}`"
                    new_obj.pop('Required')
            new_obj['Data Type'] = f"`{new_obj['Data Type']}`"
            if 'Constraints' in new_obj:
                gte = u'\N{GREATER-THAN OR EQUAL TO}'
                lte = u'\N{LESS-THAN OR EQUAL TO}'
                if type(new_obj["Constraints"]) is list:
                    new_obj["Constraints"] = ", ".join(new_obj["Constraints"])
                new_obj["Constraints"] = f"`{new_obj['Constraints'].replace('<=',lte).replace('>=',gte)}`"
            if 'Units' in new_obj:
                if new_obj['Units'] == '-':
                    new_obj['Units'] = r'\-'
                else:
                    new_obj['Units'] = new_obj['Units'].replace('-',r'·')
                    new_obj['Units'] = re.sub(r'(\d+)',r'^\1^',new_obj['Units'])
            compress_notes(new_obj)
            data_elements.append(new_obj)
        output[dat_gr] = data_elements
    return output


def enumerators_dict_from_enumerations(enumerations):
    """
    - enumerations: dict, the enumeration objects
    RETURN: list of dict, the enumeration objects as a list
    """
    output = {}
    for enum in enumerations:
        output[enum] = []
        for enumerator in enumerations[enum]["Enumerators"]:
            if enumerations[enum]["Enumerators"][enumerator]:
                item = deepcopy(enumerations[enum]["Enumerators"][enumerator])
            else:
                item = {}
            item["Enumerator"] = f"`{enumerator}`"
            compress_notes(item)
            output[enum].append(item)
    return output


def load_structure_from_object(instance):
    """
    - instance: dictionary, the result of loading a *.schema.yaml file
    RETURN: {
        'data_types': array,
        'string_types': array,
        'enumerations': dict,
        'data_groups': dict,
    }
    """
    data_types = []
    string_types = []
    enumerations = {}
    data_groups = {}

    for obj in instance:
        object_type = instance[obj]["Object Type"]
        if object_type == "Data Type":
            new_obj = instance[obj]
            new_obj["Data Type"] = f'`{obj}`'
            new_obj["Examples"] = ', '.join(new_obj["Examples"])
            data_types.append(new_obj)
        elif object_type == "String Type":
            new_obj = instance[obj]
            new_obj["String Type"] = f'`{obj}`'
            new_obj["Examples"] = ', '.join(new_obj["Examples"])
            string_types.append(new_obj)
        elif object_type == "Enumeration":
            new_obj = instance[obj]
            compress_notes(new_obj)
            enumerations[obj] = new_obj
        elif object_type == "Data Group Template":
            new_obj = instance[obj]
        elif "Data Elements" in instance[obj]:
            data_groups[obj] = instance[obj]
        elif object_type == "Meta":
            pass
        else:
            print(f"Unknown object type: {object_type}.")
    return {
        'data_types': data_types,
        'string_types': process_string_types(string_types),
        'enumerations': enumerators_dict_from_enumerations(enumerations),
        'data_groups': data_elements_dict_from_data_groups(data_groups),
    }


def trailing_ws(flag):
    """
    - flag: bool, if True, return two newlines
    RETURN: string
    """
    return "\n\n" if flag else ""


def create_table_from_list(
        columns,
        data_list,
        defaults=None,
        caption=None,
        add_training_ws=True):
    """
    - columns: array of string, the column headers
    - data_list: array of dict with keys corresponding to columns array
    - defaults: None or dict from string to value, the defaults to use for a
      column if data missing
    - caption: None or string, if specified, adds a caption
    - add_training_ws: Bool, if True, adds trailing whitespace
    RETURN: string, the table in Pandoc markdown grid table format
    """
    if len(data_list) == 0:
        return ""
    data = {col:[] for col in columns}
    for col in columns:
        data[col] = []
        for item in data_list:
            if col in item:
                data[col].append(item[col])
            elif defaults is not None and col in defaults:
                data[col].append(defaults[col])
            else:
                raise Exception(f"Expected item to have key `{col}`: `{item}`")
    return (write_table(data, columns, caption) +
            trailing_ws(add_training_ws))


def data_types_table(data_types, caption=None, add_training_ws=True):
    """
    - data_types: array of ..., the data types
    - caption: None or string, optional caption
    - add_training_ws: Bool, if True, adds trailing whitespace
    RETURN: string, the table in Pandoc markdown grid table format
    """
    return create_table_from_list(
            columns=[
                "Data Type", "Description", "JSON Schema Type", "Examples"],
            data_list=data_types,
            defaults=None,
            caption=caption,
            add_training_ws=add_training_ws)


def string_types_table(string_types, caption=None, add_training_ws=True):
    """
    - string_types: array of ..., the string types
    - caption: None or string, optional caption
    - add_training_ws: Bool, if True, adds trailing whitespace
    RETURN: string, the table in Pandoc markdown grid table format
    """
    return create_table_from_list(
            columns=[
                "String Type", "Description", "JSON Schema Pattern",
                "Examples"],
            data_list=string_types,
            caption=caption,
            add_training_ws=add_training_ws,
            defaults=None)


def enumerators_table(enumerators, caption=None, add_training_ws=True):
    """
    - enumerators: array of ..., the enumerators array
    - caption: None or string, optional caption
    - add_training_ws: Bool, if True, adds trailing whitespace
    RETURN: string, the table in Pandoc markdown grid table format
    """
    return create_table_from_list(
            columns=["Enumerator", "Description", "Notes"],
            data_list=enumerators,
            caption=caption,
            add_training_ws=add_training_ws,
            defaults={"Notes": ""})


def data_groups_table(data_elements, caption=None, add_training_ws=True):
    """
    - data_elements: array of ..., the data elements
    - caption: None or string, optional caption
    - add_training_ws: Bool, if True, adds trailing whitespace
    RETURN: string, the table in Pandoc markdown grid table format
    """
    return create_table_from_list(
            columns=[
                "Name", "Description", "Data Type", "Units", "Constraints",
                "Req", "Notes"],
            data_list=data_elements,
            caption=caption,
            add_training_ws=add_training_ws,
            defaults={"Notes": "", "Req": "", "Units": "", "Constraints": ""})

def write_data_model(instance, make_headers=False, base_level=1):
  struct = load_structure_from_object(instance)
  output = None
  with io.StringIO() as output_file:
    # Data Types
    table_type = 'data_types'
    if len(struct[table_type]) > 0:
      table_title = "Data Types"
      if make_headers:
        output_file.writelines(write_header(table_title, base_level))
        caption = None
      else:
        caption = table_title
      output_file.writelines(data_types_table(struct[table_type], caption=caption))
    # String Types
    table_type = 'string_types'
    if len(struct[table_type]) > 0:
      table_title = "String Types"
      if make_headers:
        output_file.writelines(write_header(table_title, base_level))
        caption = None
      else:
        caption = table_title
      output_file.writelines(string_types_table(struct[table_type], caption=caption))
    # Enumerations
    output_file.writelines(write_header("Enumerations", base_level))
    table_type = 'enumerations'
    if len(struct[table_type]) > 0:
      for enum, enumerators in struct[table_type].items():
        table_title = enum
        if make_headers:
            output_file.writelines(write_header(table_title, base_level + 1))
            caption = None
        else:
            caption = table_title
        output_file.writelines(enumerators_table(enumerators, caption=caption))
    else:
      output_file.writelines(["None.","\n"*2])
    # Data Groups
    output_file.writelines(write_header("Data Groups", base_level))
    table_type = 'data_groups'
    if len(struct[table_type]) > 0:
      for dg, data_elements in struct[table_type].items():
        table_title = dg
        if make_headers:
            output_file.writelines(write_header(table_title, base_level + 1))
            caption = None
        else:
            caption = table_title
        output_file.writelines(data_groups_table(data_elements, caption=caption))
    else:
      output_file.writelines(["None.","\n"*2])
    output = output_file.getvalue()
  return output



import os
import json
import cbor2
import yaml
import shutil

def get_extension(file):
    return os.path.splitext(file)[1]

def get_file_basename(file, depth=0):
    basename = os.path.basename(file)
    for i in range(depth):
        basename = os.path.splitext(basename)[0]
    return basename

def load(input_file_path):
    ext = get_extension(input_file_path).lower()
    if (ext == '.json'):
        with open(input_file_path, 'r') as input_file:
            return json.load(input_file)
    elif (ext == '.cbor'):
        with open(input_file_path, 'rb') as input_file:
            return cbor2.load(input_file)
    elif (ext == '.yaml') or (ext == '.yml'):
        with open(input_file_path, 'r') as input_file:
            return yaml.load(input_file, Loader=yaml.FullLoader)
    else:
        raise Exception(f"Unsupported input \"{ext}\".")

def dump(content, output_file_path):
    ext = get_extension(output_file_path).lower()
    if (ext == '.json'):
        with open(output_file_path,'w') as output_file:
            json.dump(content, output_file, indent=4)
    elif (ext == '.cbor'):
        with open(output_file_path,'wb') as output_file:
            cbor2.dump(content, output_file)
    elif (ext == '.yaml') or (ext == '.yml'):
        with open(output_file_path, 'w') as out_file:
            yaml.dump(content, out_file, sort_keys=False)

    else:
        raise Exception(f"Unsupported output \"{ext}\".")

def dump_to_string(content, output_type='json'):
    if (output_type == 'json'):
        return json.dump(content, indent=4)
    elif (output_type == 'yaml') or (output_type == 'yml'):
        return yaml.dump(content, sort_keys=False)
    else:
        raise Exception(f"Unsupported output \"{output_type}\".")

def translate(input, output):
    dump(load(input),output)

def make_dir(dir_path):
  if not os.path.exists(dir_path):
      os.mkdir(dir_path)
  return dir_path

def remove_dir(dir_path):
    if os.path.exists(dir_path) and os.path.isdir(dir_path):
        shutil.rmtree(dir_path)

def check_dir(dir_path, dir_description="Directory"):
    if not os.path.exists(dir_path):
      raise Exception(f"{dir_description}, \"{dir_path}\", does not exist.")
    elif not os.path.isdir(dir_path):
      raise Exception(f"{dir_description}, \"{dir_path}\", is not a directory.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lattice', 'lattice.docs']

package_data = \
{'': ['*'],
 'lattice.docs': ['hugo_layouts/landing/*',
                  'hugo_layouts/partials/*',
                  'hugo_layouts/specifications/*']}

install_requires = \
['GitPython==3.1.18',
 'Jinja2',
 'cbor2',
 'jsonschema',
 'pyyaml',
 'stringcase==1.2.0']

setup_kwargs = {
    'name': 'lattice',
    'version': '0.1.0',
    'description': '',
    'long_description': 'None',
    'author': 'Big Ladder Software',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .file_io import load, dump

import os
import re
import jsonschema
import posixpath

# -------------------------------------------------------------------------------------------------
class DataGroup:

    # The following regular expressions are explicity not Unicode
    array_type = r'^\[(.*?)\]'                       # Array of type 'Type' e.g. '[Type]'
                                                    # just the first [] pair; using non-greedy '?'
    minmax_range_type = r'^(?P<min>[0-9]*)(?P<ellipsis>\.*)(?P<max>[0-9]*)'    # Parse ellipsis range-notation e.g. '[1..]'
    alternative_type = r'^\((.*)\)'                  # Parentheses encapsulate a list of options
    enum_or_def = r'(\{|\<)(.*)(\}|\>)'
    numeric_type = r'[+-]?[0-9]*\.?[0-9]+|[0-9]+'   # Any optionally signed, floating point number
    scope_constraint = r'^:(.*):'                    # Lattice scope constraint for ID/Reference

    def __init__(self, name, type_list, ref_list=None, **kwargs):
        self._name = name
        self._types = type_list
        self._refs = ref_list

    def add_data_group(self, group_name, group_subdict):
        '''
        Process Data Group from the source schema into a properties node in json.

        :param group_name:      Data Group name; this will become a schema definition key
        :param group_subdict:   Dictionary of Data Elements where each element is a key
        '''
        elements = {'type': 'object',
                    'properties' : dict()}
        required = list()
        dependencies = dict()
        for e in group_subdict:
            element = group_subdict[e]
            if 'Description' in element:
                elements['properties'][e] = {'description' : element['Description']}
            if 'Data Type' in element:
                self._create_type_entry(group_subdict[e], elements, e)
            if 'Units' in element:
                elements['properties'][e]['units'] = element['Units']
            if 'Notes' in element:
                elements['properties'][e]['notes'] = element['Notes']
            if 'Required' in element:
                req = element['Required']
                if isinstance(req, bool):
                    if req == True:
                        required.append(e)
                elif req.startswith('if'):
                    self._construct_requirement_if_then(elements, dependencies, req[3:], e)
        if required:
            elements['required'] = required
        if dependencies:
            elements['dependencies'] = dependencies
        elements['additionalProperties'] = False
        return {group_name : elements}


    def _construct_requirement_if_then(self,
                                       conditionals_list : dict,
                                       dependencies_list : dict,
                                       requirement_str : str,
                                       requirement : str):
        '''
        Construct paired if-then json entries for conditional requirements.

        :param target_list_to_append:   List of dictionaries, modified in-situ with an if key and
                                        an associated then key
        :param requirement_str:         Raw requirement string using A205 syntax
        :param requirement:             This item's presence is dependent on the above condition
        '''
        separator = r'\sand\s'
        collector = 'allOf'
        selector_dict = {'properties' : {collector : dict()}}
        requirement_list = re.split(separator, requirement_str)
        dependent_req = r'(?P<selector>!?[0-9a-zA-Z_]*)((?P<is_equal>!?=)(?P<selector_state>[0-9a-zA-Z_]*))?'

        for req in requirement_list:
            m = re.match(dependent_req, req)
            if m:
                selector = m.group('selector')
                if m.group('is_equal'):
                    is_equal = False if '!' in m.group('is_equal') else True
                    selector_state = m.group('selector_state')
                    if 'true' in selector_state.lower():
                        selector_state = True
                    elif 'false' in selector_state.lower():
                        selector_state = False
                    selector_dict['properties'][collector][selector] = {'const' : selector_state} if is_equal else {'not' : {'const' : selector_state} }
                else: # prerequisite type
                    if dependencies_list.get(selector):
                        dependencies_list[selector].append(requirement)
                    else:
                        if '!' in selector:
                            dependencies_list[selector.lstrip('!')] = {'not' : {'required' : [requirement]}}
                        else:
                            dependencies_list[selector] = [requirement]

        if selector_dict['properties'][collector].keys():
            # Conditional requirements are each a member of a list
            if conditionals_list.get('allOf') == None:
                conditionals_list['allOf'] = list()

            for conditional_req in conditionals_list['allOf']:
                if conditional_req.get('if') == selector_dict: # condition already exists
                    conditional_req['then']['required'].append(requirement)
                    return
            conditionals_list['allOf'].append(dict())
            conditionals_list['allOf'][-1]['if'] = selector_dict
            conditionals_list['allOf'][-1]['then'] = {'required' : [requirement]}


    def _create_type_entry(self, parent_dict, target_dict, entry_name):
        '''
        Create json type node and its nested nodes if necessary.
        :param parent_dict:     A Data Element's subdictionary, from source schema
        :param target_dict:     The json definition node that will be populated
        :param entry_name:      Data Element name
        '''
        m = re.findall(DataGroup.array_type, parent_dict['Data Type'])
        if entry_name not in target_dict['properties']:
            target_dict['properties'][entry_name] = dict()
        target_property_entry = target_dict['properties'][entry_name]
        if m: # Data Element is an array
            # 1. 'type' entry
            target_property_entry['type'] = 'array'
            # 2. 'm[in/ax]Items' entry
            if len(m) > 1:
                # Parse ellipsis range-notation e.g. '[1..]'
                mnmx = re.match(DataGroup.minmax_range_type, m[1])
                target_property_entry['minItems'] = int(mnmx.group('min'))
                if (mnmx.group('ellipsis') and mnmx.group('max')):
                    target_property_entry['maxItems'] = int(mnmx.group('max'))
                elif not mnmx.group(2):
                    target_property_entry['maxItems'] = int(mnmx.group('min'))
            # 3. 'items' entry
            target_property_entry['items'] = dict()
            self._get_simple_type(m[0], target_property_entry['items'])
            self._get_numeric_constraints(parent_dict.get('Constraints'), target_property_entry['items'])
        else:
            # If the type is oneOf a set
            m = re.match(DataGroup.alternative_type, parent_dict['Data Type'])
            if m:
                types = [t.strip() for t in m.group(1).split(',')]
                selection_key, selections = parent_dict['Constraints'].split('(')
                target_dict['allOf'] = list()
                for s, t in zip(selections.split(','), types):
                    target_dict['allOf'].append(dict())
                    self._construct_selection_if_then(target_dict['allOf'][-1], selection_key, s, entry_name)
                    self._get_simple_type(t, target_dict['allOf'][-1]['then']['properties'][entry_name])
            elif parent_dict['Data Type'] in ['ID', 'Reference']:
                m = re.match(DataGroup.scope_constraint, parent_dict['Constraints'])
                if m:
                    target_property_entry['scopedType'] = parent_dict['Data Type']
                    target_property_entry['scope'] = m.group(1)
                    self._get_simple_type(parent_dict['Data Type'], target_property_entry)
            else:
                # 1. 'type' entry
                self._get_simple_type(parent_dict['Data Type'], target_property_entry)
                # 2. 'm[in/ax]imum' entry
                self._get_numeric_constraints(parent_dict.get('Constraints'), target_property_entry)


    def _construct_selection_if_then(self, target_dict_to_append, selector, selection, entry_name):
        '''
        Construct paired if-then json entries for allOf collections translated from source-schema
        "selector" Constraints.

        :param target_dict_to_append:   This dictionary is modified in-situ with an if key and
                                        associated then key
        :param selector:                Constraints key
        :param selection:               Item from constraints values list.
        :param entry_name:              Data Element for which the Data Type must match the
                                        Constraint
        '''
        target_dict_to_append['if'] = {'properties' : {selector : {'const' : ''.join(ch for ch in selection if ch.isalnum())} } }
        target_dict_to_append['then'] = {'properties' : {entry_name : dict()}}


    def _get_simple_type(self, type_str, target_dict_to_append):
        ''' Return the internal type described by type_str, along with its json-appropriate key.
            First, attempt to capture enum, definition, or special string type as references;
            then default to fundamental types with simple key "type".

            :param type_str:                Input string from source schema's Data Type key
            :param target_dict_to_append:   The json "items" node
        '''
        internal_type = None
        m = re.match(DataGroup.enum_or_def, type_str)
        if m:
            internal_type = m.group(2)
        else:
            internal_type = type_str
        # Look through the references to assign a source to the type
        for key in self._refs:
            if internal_type in self._refs[key]:
                internal_type = key + '.schema.json#/definitions/' + internal_type
                target_dict_to_append['$ref'] = internal_type
                return
        try:
            if '/' in type_str:
                # e.g. "Numeric/Null" becomes a list of 'type's
                target_dict_to_append['type'] = [self._types[t] for t in type_str.split('/')]
            else:
                target_dict_to_append['type'] = self._types[type_str]
        except KeyError:
            print('Type not processed:', type_str)
        return


    def _get_numeric_constraints(self, constraints_str, target_dict):
        '''
        Process numeric Constraints into fields.

        :param constraints_str:     Raw numerical limits and/or multiple information
        :param target_dict:         json property node
        '''
        if constraints_str is not None:
            constraints = constraints_str if isinstance(constraints_str, list) else [constraints_str]
            minimum=None
            maximum=None
            for c in constraints:
                try:
                    # Process numeric constraints
                    numerical_value = re.findall(DataGroup.numeric_type, c)[0]
                    if '>' in c:
                        minimum = (float(numerical_value) if 'number' in target_dict['type'] else int(numerical_value))
                        mn = 'exclusiveMinimum' if '=' not in c else 'minimum'
                        target_dict[mn] = minimum
                    elif '<' in c:
                        maximum = (float(numerical_value) if 'number' in target_dict['type']  else int(numerical_value))
                        mx = 'exclusiveMaximum' if '=' not in c else 'maximum'
                        target_dict[mx] = maximum
                    elif '%' in c:
                        target_dict['multipleOf'] = int(numerical_value)
                except IndexError:
                    pass
                except ValueError:
                    pass
                except KeyError:
                    # 'type' not in dictionary
                    pass


# -------------------------------------------------------------------------------------------------
class Enumeration:

    def __init__(self, name, description=None):
        self._name = name
        self._enumerants = list() # list of tuple:[value, description, display_text, notes]
        self.entry = dict()
        self.entry[self._name] = dict()
        if description:
            self.entry[self._name]['description'] = description

    def add_enumerator(self, value, description=None, display_text=None, notes=None):
        '''Store information grouped as a tuple per enumerant.'''
        self._enumerants.append((value, description, display_text, notes))

    def create_dictionary_entry(self):
        '''
        Convert information currently grouped per enumerant, into json groups for
        the whole enumeration.
        '''
        z = list(zip(*self._enumerants))
        enums = {'type': 'string',
                 'enum' : z[0]}
        if any(z[2]):
            enums['enum_text'] = z[2]
        if any(z[1]):
            enums['descriptions'] = z[1]
        if any(z[3]):
            enums['notes'] = z[3]
        self.entry[self._name] = {**self.entry[self._name], **enums}
        return self.entry


# -------------------------------------------------------------------------------------------------
class JSON_translator:
    def __init__(self, **kwargs):
        ''' '''
        self._references = dict()
        self._fundamental_data_types = dict()
        self._schema_object_types = ['Data Group', 'String Type', 'Enumeration'] # "Basic" object types - are there more?
        self._kwargs = kwargs

    def load_source_schema(self, input_file_path):
        '''Load and process a yaml schema into its json schema equivalent.'''
        self._schema = {'$schema': 'http://json-schema.org/draft-07/schema#',
                        'title': None,
                        'description': None,
                        'definitions' : dict()}
        self._references.clear()
        self._source_dir = os.path.dirname(os.path.abspath(input_file_path))
        self._schema_name = os.path.splitext(os.path.splitext(os.path.basename(input_file_path))[0])[0]
        self._fundamental_data_types.clear()
        self._contents = load(input_file_path)
        sch = dict()
        # Iterate through the dictionary, looking for known types
        for base_level_tag in self._contents:
            if 'Object Type' in self._contents[base_level_tag]:
                obj_type = self._contents[base_level_tag]['Object Type']
                if obj_type == 'Meta':
                    self._load_meta_info(self._contents[base_level_tag])
                elif obj_type == 'String Type':
                    if 'Is Regex' in self._contents[base_level_tag]:
                        sch.update({base_level_tag : {"type":"string", "regex":True}})
                    else:
                        sch.update({base_level_tag : {"type":"string", "pattern":self._contents[base_level_tag]['JSON Schema Pattern']}})
                elif obj_type == 'Enumeration':
                    sch.update(self._process_enumeration(base_level_tag))
                elif obj_type in self._schema_object_types:
                    dg = DataGroup(base_level_tag, self._fundamental_data_types, self._references)
                    sch.update(dg.add_data_group(base_level_tag, self._contents[base_level_tag]['Data Elements']))
        self._schema['definitions'] = sch
        return self._schema

    def _load_meta_info(self, schema_section):
        '''Store the global/common types and the types defined by any named references.'''
        self._schema['title'] = schema_section['Title']
        self._schema['description'] = schema_section['Description']
        if 'Version' in schema_section:
            self._schema['version'] = schema_section['Version']
        if 'Root Data Group' in schema_section:
            self._schema['$ref'] = self._schema_name + '.schema.json#/definitions/' + schema_section['Root Data Group']
        # Create a dictionary of available external objects for reference
        refs = {'core' : os.path.join(os.path.dirname(__file__),'core.schema.yaml'),
                f'{self._schema_name}' : os.path.join(self._source_dir, f'{self._schema_name}.schema.yaml')}
        if 'References' in schema_section:
            for ref in schema_section['References']:
                refs.update({f'{ref}' : os.path.join(self._source_dir, ref + '.schema.yaml')})
        for ref_name in refs:
            try:
                ext_dict = load(refs[ref_name])
                # Only one of the references should contain Data Group Templates
                for template_name in [ext_dict[name]['Name'] for name in ext_dict if ext_dict[name]['Object Type'] == 'Data Group Template']:
                    self._schema_object_types.append(template_name)
                # Template data groups may exist in any of the references
                self._references[ref_name] = [base_item for base_item in ext_dict if ext_dict[base_item]['Object Type'] in self._schema_object_types]
                for base_item in [name for name in ext_dict if ext_dict[name]['Object Type'] == 'Data Type']:
                    self._fundamental_data_types[base_item] = ext_dict[base_item]['JSON Schema Type']
            except RuntimeError:
                raise RuntimeError(f'{refs[ref_name]} reference file does not exist.') from None

    def _process_enumeration(self, name_key):
        ''' Collect all Enumerators in an Enumeration block. '''
        enums = self._contents[name_key]['Enumerators']
        description = self._contents[name_key].get('Description')
        definition = Enumeration(name_key, description)
        for key in enums:
            try:
                descr = enums[key]['Description']  if 'Description'  in enums[key] else None
                displ = enums[key]['Display Text'] if 'Display Text' in enums[key] else None
                notes = enums[key]['Notes']        if 'Notes'        in enums[key] else None
                definition.add_enumerator(key, descr, displ, notes)
            except TypeError: # key's value is None
                definition.add_enumerator(key)
        return definition.create_dictionary_entry()


# -------------------------------------------------------------------------------------------------
class JSONSchemaValidator:
    def __init__(self, schema_path):
        with open(schema_path) as schema_file:
            uri_path = os.path.abspath(os.path.dirname(schema_path))
            if os.sep != posixpath.sep:
                uri_path = posixpath.sep + uri_path
            resolver = jsonschema.RefResolver(f'file://{uri_path}/', schema_file)
            self.validator = jsonschema.Draft7Validator(load(schema_path), resolver=resolver)

    def validate(self, instance_path):
        instance = load(instance_path)
        errors = sorted(self.validator.iter_errors(instance), key=lambda e: e.path)
        file_name =  os.path.basename(instance_path)
        if len(errors) == 0:
            print(f"Validation successful for {file_name}")
        else:
            messages = []
            for error in errors:
                messages.append(f"{error.message} ({'.'.join([str(x) for x in error.path])})")
            messages = [f"{i}. {message}" for i, message in enumerate(messages, start=1)]
            message_str = '\n  '.join(messages)
            raise Exception(f"Validation failed for {file_name} with {len(messages)} errors:\n  {message_str}")

# -------------------------------------------------------------------------------------------------
def generate_core_json_schema():
    '''Create JSON schema from core YAML schema'''
    j = JSON_translator()
    core_instance = j.load_source_schema(os.path.join(os.path.dirname(__file__),'core.schema.yaml'))
    return core_instance

# -------------------------------------------------------------------------------------------------
def search_for_reference(referenced_schemas: dict, schema_path: str, subdict: dict) -> bool:
    '''Search for $ref keys and replace the associated dictionary entry in-situ.'''
    subbed = False
    if '$ref' in subdict.keys():
        subbed = True
        # parse the ref and locate the sub-dict
        source_file, ref_loc = subdict['$ref'].split('#')
        ref = referenced_schemas[os.path.splitext(source_file)[0]]
        key_tree = ref_loc.lstrip('/').split('/')
        sub_d = ref[key_tree[0]]
        for k in key_tree[1:]:
            sub_d = sub_d[k]
        # replace the $ref with its contents
        subdict.update(sub_d)
        subdict.pop('$ref')
        # re-search the substituted dictionary
        subbed = search_for_reference(referenced_schemas, schema_path, subdict)
    else:
        for key in subdict:
            if isinstance(subdict[key], dict):
                subbed = search_for_reference(referenced_schemas, schema_path, subdict[key])
            if isinstance(subdict[key], list):
                for entry in [item for item in subdict[key] if isinstance(item, dict)]:
                    subbed = search_for_reference(referenced_schemas, schema_path, entry)
    return subbed

# -------------------------------------------------------------------------------------------------
def generate_json_schema(source_schema_input_path, json_schema_output_path):
    '''Create reference-resolved JSON schema from YAML source schema.'''
    if os.path.isfile(source_schema_input_path) and '.schema.yaml' in source_schema_input_path:
        schema_dir = os.path.abspath(os.path.dirname(source_schema_input_path))
        json_schema_output_dir = os.path.abspath(os.path.dirname(json_schema_output_path))
        j = JSON_translator()
        main_schema_instance = j.load_source_schema(source_schema_input_path)
        schema_ref_map = {'core.schema' : generate_core_json_schema()}
        for ref_source in os.listdir(schema_dir):
            schema_ref_map[os.path.splitext(ref_source)[0]] = j.load_source_schema(os.path.join(schema_dir, ref_source))
        while search_for_reference(schema_ref_map, schema_dir, main_schema_instance):
            pass
        dump(main_schema_instance, json_schema_output_path)

# -------------------------------------------------------------------------------------------------
def get_scope_locations(schema: dict, scopes_dict: dict, scope_key: str='Reference', lineage: list=None):
    if not lineage:
        lineage = list()
    for key in schema:
        if key == 'scopedType' and schema[key] == scope_key:
            scopes_dict['.'.join(lineage)] = schema['scope'] # key is a dot-separated path
        elif isinstance(schema[key], dict):
            get_scope_locations(schema[key], scopes_dict, scope_key, lineage + [key])
        elif isinstance(schema[key], list):
            for entry in [item for item in schema[key] if isinstance(item, dict)]:
                get_scope_locations(entry, scopes_dict, scope_key, lineage + [key])

# -------------------------------------------------------------------------------------------------
def get_reference_value(data_dict: dict, lineage: list) -> str:
    test_reference = data_dict
    for adr in lineage:
        if test_reference.get(adr):
            if isinstance(test_reference[adr], list):
                for item in test_reference[adr]:
                    return get_reference_value(item, lineage[1:])
            else:
                test_reference = test_reference[adr]
        else:
            return None
    return test_reference

# -------------------------------------------------------------------------------------------------
def postvalidate_references(input_file, input_schema):
    id_scopes = dict()
    reference_scopes = dict()
    get_scope_locations(load(input_schema), scope_key='ID', scopes_dict=id_scopes)
    get_scope_locations(load(input_schema), scope_key='Reference', scopes_dict=reference_scopes)
    data = load(input_file)
    ids = list()
    for id_loc in [id for id in id_scopes if id.startswith('properties')]:
        lineage = [level for level in id_loc.split('.') if level not in ['properties', 'items']]
        ids.append(get_reference_value(data, lineage))
    for ref in [r for r in reference_scopes if r.startswith('properties')]:
        lineage = [level for level in ref.split('.') if level not in ['properties', 'items']]
        reference_scope = get_reference_value(data, lineage)
        if reference_scope != None and reference_scope not in ids:
            raise Exception(f'Scope mismatch in {input_file}; {reference_scope} not in ID scope list {ids}.')

# -------------------------------------------------------------------------------------------------
def validate_file(input_file, input_schema):
    v = JSONSchemaValidator(input_schema)
    v.validate(input_file)

# -------------------------------------------------------------------------------------------------
def postvalidate_file(input_file, input_schema):
    postvalidate_references(input_file, input_schema) 
